/*
 * Copyright (C) 2007 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: aabag.c,v 1.8 2007/06/20 21:27:27 nalin Exp $"

#include "../config.h"

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <fnmatch.h>
#include <limits.h>
#include <stdio.h>

#include <nspr.h>
#include <nss.h>
#include <keyhi.h>
#include <pk11pub.h>
#include <secport.h>

#include "aabag.h"
#include "bpk5.h"
#include "commont.h"
#include "pkinit.h"

#define WHITESPACE " \t"
#define PKINIT_CA_TRUST_FLAGS (CERTDB_VALID_PEER | \
			       CERTDB_VALID_CA | \
			       CERTDB_TRUSTED_CA | \
			       CERTDB_TRUSTED_CLIENT_CA)
#define PKINIT_CA_VALID_FLAGS (CERTDB_VALID_CA)
#define PKINIT_KEY_USAGE_FLAGS (KU_DIGITAL_SIGNATURE | \
				KU_KEY_ENCIPHERMENT | \
				KU_DATA_ENCIPHERMENT)

struct aa_item {
	char *item_type;
	SECItem *item, *ciphertext;
	struct aa_item *next;
};

struct aa_dir {
	const char *path;
	struct aa_dir *next;
};

struct aa_key_list {
	struct aa_key_list *next;
	SECItem *item;
};

struct aa_bag {
	struct module_context *mcontext;
	PK11PasswordFunc pwcb;
	PLArenaPool *bag_pool;
	CERTCertDBHandle *certdb;
	SECKEYPrivateKeyList *klist;
	CERTCertList *clist;
	struct aa_key_list *key_items;
	struct aa_item *encrypted_items;
	struct aa_dir *cert_dirs, *cacert_dirs;
};

static const signed char
b64[96] = {
	/* 0x20 */	   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
	/* 0x28 */	   -1,   -1,   -1,   62,   -1,   -1,   -1,   63, 
	/* 0x30 */	   52,   53,   54,   55,   56,   57,   58,   59, 
	/* 0x38 */	   60,   61,   -1,   -1,   -1,   -1,   -1,   -1, 
	/* 0x40 */	   -1,    0,    1,    2,    3,    4,    5,    6, 
	/* 0x48 */	    7,    8,    9,   10,   11,   12,   13,   14, 
	/* 0x50 */	   15,   16,   17,   18,   19,   20,   21,   22, 
	/* 0x58 */	   23,   24,   25,   -1,   -1,   -1,   -1,   -1, 
	/* 0x60 */	   -1,   26,   27,   28,   29,   30,   31,   32, 
	/* 0x68 */	   33,   34,   35,   36,   37,   38,   39,   40, 
	/* 0x70 */	   41,   42,   43,   44,   45,   46,   47,   48, 
	/* 0x78 */	   49,   50,   51,   52,   -1,   -1,   -1,   -1
};

static long
unb64(const unsigned char *base64, int len, unsigned char *out)
{
	int i, parts, acc;
	long j;
	unsigned char tmp;
	parts = 0;
	j = 0;
	acc = 0;
	for (i = 0; i < len; i++) {
		tmp = base64[i] & 0xff;
		if ((tmp < 0x20) || (tmp > 0x7f)) {
			continue;
		}
		if (tmp == '=') {
			if (parts >= 2) {
				out[j++] = (acc >> 16) & 0xff;
				if (parts >= 3) {
					out[j++] = (acc >>  8) & 0xff;
					if (parts >= 4) {
						out[j++] = (acc >>  0) & 0xff;
					}
				}
			}
			break;
		}
		switch (parts) {
		case 0:
			acc |= (b64[tmp - 0x20] << 18);
			parts++;
			break;
		case 1:
			acc |= (b64[tmp - 0x20] << 12);
			parts++;
			break;
		case 2:
			acc |= (b64[tmp - 0x20] << 6);
			parts++;
			break;
		case 3:
			acc |= (b64[tmp - 0x20] << 0);
			parts++;
			break;
		}
		if (parts == 4) {
			out[j++] = (acc >> 16) & 0xff;
			out[j++] = (acc >>  8) & 0xff;
			out[j++] = (acc >>  0) & 0xff;
			parts = 0;
			acc = 0;
		}
	}
	return j;
}

static long
unhex(const char *hex, int len, unsigned char *out)
{
	int acc, parts, i;
	long j;
	const char hexchars[] = "00112233445566778899aAbBcCdDeEfF", *p;
	acc = 0;
	parts = 0;
	j = 0;
	for (i = 0; i < len; i++) {
		p = strchr(hexchars, hex[i]);
		if (p != NULL) {
			acc = (acc << 4) | ((p - hexchars) / 2);
			parts++;
			if (parts == 2) {
				out[j++] = acc & 0xff;
				acc = 0;
				parts = 0;
			}
		}
	}
	return j;
}

static void
fix_des_parity(SECItem *item)
{
	unsigned int i;
	unsigned char c;
	for (i = 0; i < item->len; i++) {
		c = item->data[i];
		c = c & 0xfe;
		c |= ((c >> 6) & 1) ^
		     ((c >> 5) & 1) ^
		     ((c >> 4) & 1) ^
		     ((c >> 3) & 1) ^
		     ((c >> 2) & 1) ^
		     ((c >> 1) & 1) ^
		     ((c >> 0) & 1);
		item->data[i] = c;
	}
}

static SECItem *
decrypt_item(struct module_context *mcontext, PLArenaPool *pool,
	     const char *alg, const char *param, SECItem *item,
	     PK11PasswordFunc pwcb, struct pkinit_pwcb_args *pwcb_args)
{
	CK_MECHANISM_TYPE mech;
	SECItem *pass, *iv, *key, *params, *ret;
	PK11SlotInfo *slot;
	PK11SymKey *symkey;
	PK11Context *ctx;
	unsigned char *buf, pad, blocksize;
	char *password;
	int buflen, keylen;

	params = NULL;
	if (strcmp(alg, "DES-CBC") == 0) {
		mech = CKM_DES_CBC;
		keylen = 8;
		blocksize = 8;
	} else
	if (strcmp(alg, "DES-EDE3-CBC") == 0) {
		mech = CKM_DES3_CBC;
		keylen = 24;
		blocksize = 8;
	} else
	if (strcmp(alg, "AES-128-CBC") == 0) {
		mech = CKM_AES_CBC;
		keylen = 16;
		blocksize = 16;
	} else
	if (strcmp(alg, "AES-192-CBC") == 0) {
		mech = CKM_AES_CBC;
		keylen = 24;
		blocksize = 16;
	} else
	if (strcmp(alg, "AES-256-CBC") == 0) {
		mech = CKM_AES_CBC;
		keylen = 32;
		blocksize = 16;
	} else {
		pkinit_debug(mcontext, 1, "Unrecognized PEM cipher \"%s\".\n",
			     alg);
		mech = 0;
		keylen = 0;
		blocksize = 0;
		return NULL;
	}

	/* We actually need a callback now. */
	if (pwcb == NULL) {
		pkinit_debug(mcontext, 0,
			     "Unable to prompt for password.\n");
		return NULL;
	}

	/* Get a handle on the internal key slot. */
	slot = PK11_GetBestSlot(mech, pwcb_args);
	if (slot == NULL) {
		pkinit_debug(mcontext, 0,
			     "Error locating internal key slot.\n");
		return NULL;
	}

	/* Get the password and generate the key. */
	password = pwcb(slot, PR_FALSE, pwcb_args);
	if (password == NULL) {
		pkinit_debug(mcontext, 0,
			     "Error reading password.\n");
		PK11_FreeSlot(slot);
		return NULL;
	}

	pass = SECITEM_AllocItem(pool, NULL, strlen(password));
	memcpy(pass->data, password, pass->len);
	memset(password, 0, strlen(password));
	PR_Free(password);

	iv = SECITEM_AllocItem(pool, NULL, strlen(param));
	iv->len = unhex(param, strlen(param), iv->data);
	if (iv->len > 8) {
		iv->len = 8;
	}

	if (bpk5_pbkdf1(mcontext, pool, pass, iv, 1, SEC_OID_MD5,
			keylen, &key) != SECSuccess) {
		pkinit_debug(mcontext, 1, "Unrecognized PEM cipher \"%s\".\n",
			     alg);
		PK11_FreeSlot(slot);
		return NULL;

	}
	/* Fix parity on the key byte for DES-based ciphers. */
	switch (mech) {
	case CKM_DES_CBC:
	case CKM_DES3_CBC:
		fix_des_parity(key);
		break;
	case CKM_AES_CBC:
	default:
		break;
	}

	/* Import the key. */
	symkey = PK11_ImportSymKey(slot, mech,
				   PK11_OriginGenerated, CKA_DECRYPT,
				   key, pwcb_args);
	if (symkey == NULL) {
		pkinit_debug(mcontext, 0,
			     "Error preparing password-based symmetric key.\n");
		PK11_FreeSlot(slot);
		return NULL;
	}

	/* For DES/AES ciphers, the parameters are the raw IV. */
	params = iv;
	ctx = PK11_CreateContextBySymKey(mech, CKA_DECRYPT, symkey, params);
	if (ctx == NULL) {
		pkinit_debug(mcontext, 0,
			     "Error preparing to decrypt data.\n");
		PK11_FreeSymKey(symkey);
		PK11_FreeSlot(slot);
		return NULL;
	}
	buf = PORT_ZAlloc(item->len);
	if (buf == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		PK11_DestroyContext(ctx, PR_TRUE);
		PK11_FreeSymKey(symkey);
		PK11_FreeSlot(slot);
		return NULL;
	}
	if (PK11_CipherOp(ctx, buf, &buflen, item->len,
			  item->data, item->len) != SECSuccess) {
		pkinit_debug(mcontext, 0, "Error decrypting item.\n");
		PORT_Free(buf);
		PK11_DestroyContext(ctx, PR_TRUE);
		PK11_FreeSymKey(symkey);
		PK11_FreeSlot(slot);
		return NULL;
	}
	PK11_Finalize(ctx);
	PK11_DestroyContext(ctx, PR_TRUE);
	PK11_FreeSymKey(symkey);
	if (buflen > blocksize) {
		pad = (unsigned char) buf[buflen - 1];
		if (pad > blocksize) {
			pkinit_debug(mcontext, 1, "Error decrypting item: "
				     "padding check failed.\n");
			PORT_Free(buf);
			PK11_FreeSlot(slot);
			return NULL;
		}
		buflen -= pad;
	}
	ret = SECITEM_AllocItem(pool, NULL, buflen);
	if (ret != NULL) {
		memcpy(ret->data, buf, buflen);
		ret->len = buflen;
	}
	PORT_Free(buf);
	PK11_FreeSlot(slot);

	return ret;
}

static struct aa_item *
aa_item_copy(struct module_context *mcontext, PLArenaPool *pool,
	     struct aa_item *item) 
{
	struct aa_item *ret;
	ret = PORT_ArenaZAlloc(pool, sizeof(*ret));
	if (ret != NULL) {
		ret->item = SECITEM_ArenaDupItem(pool, item->item);
		if (ret->item == NULL) {
			ret = NULL;
		} else {
			ret->item_type = PORT_ArenaStrdup(pool,
							  item->item_type);
			if (ret->item_type == NULL) {
				ret = NULL;
			}
		}
		if ((ret != NULL) && (item->ciphertext != NULL)) {
			ret->ciphertext = SECITEM_ArenaDupItem(pool,
							       item->ciphertext);
			if (ret->ciphertext == NULL) {
				ret = NULL;
			}
		}
	}
	return ret;
}

static PRBool
aa_item_in_ciphertext_list(struct aa_item *list, const SECItem *item,
			   struct aa_item **match)
{
	*match = NULL;
	while (list != NULL) {
		if ((list->ciphertext != NULL) &&
		    (list->ciphertext->len == item->len) &&
		    (memcmp(list->ciphertext->data, item->data,
			    list->ciphertext->len) == 0)) {
			*match = list;
			return PR_TRUE;
		}
		list = list->next;
	}
	return PR_FALSE;
}

struct aa_item *
aa_item_from_fp(struct module_context *mcontext, PLArenaPool *pool,
		struct aa_bag *bag, FILE *fp,
		PK11PasswordFunc pwcb, struct pkinit_pwcb_args *pwcb_args)
{
	struct aa_item *item, *tmpitem;
	SECItem *plaintext;
	char line[LINE_MAX], *type, *disp, *alg, *params;
	char *p, *q;
	int headers, len;
	long version;
	char *buffer, *tmp, *raw;
	unsigned long buflen;

	if (fp == NULL) {
		return NULL;
	}
	/* Skip to the BEGIN line and extract the type. */
	do {
		if (fgets(line, sizeof(line), fp) == NULL) {
			return NULL;
		}
		len = strcspn(line, "\r\n");
		line[len] = '\0';
		if (len < 16) {
			continue;
		}
	} while ((strncmp(line, "-----BEGIN ", 11) != 0) ||
		 (strcmp(line + len - 5, "-----") != 0));
	type = PORT_ArenaStrdup(pool, line + 11);
	type[len - 11 - 5] = '\0';
	pkinit_debug(mcontext, 3, "Reading a \"%s\".\n", type);
	/* Now read each line until we find a matching end marker. */
	buffer = NULL;
	buflen = 0;
	for (;;) {
		if (fgets(line, sizeof(line), fp) == NULL) {
			return NULL;
		}
		len = strcspn(line, "\r\n");
		line[len] = '\0';
		while ((len > 0) &&
		       (strpbrk(line + (len - 1), WHITESPACE) != NULL)) {
			len--;
			line[len] = '\0';
		}
		if ((strncmp(line, "-----END ", 9) == 0) &&
		    (strncmp(line + 9, type, len - 14) == 0) &&
		    (strcmp(line + len - 5, "-----") == 0)) {
			break;
		}
		tmp = PORT_Realloc(buffer, buflen + len + 2);
		if (tmp == NULL) {
			PORT_Free(buffer);
			pkinit_debug(mcontext, 0, "Out of memory.\n");
			buffer = NULL;
			break;
		}
		buffer = tmp;
		if (buflen > 0) {
			buffer[buflen] = '\n';
			buflen++;
			memcpy(buffer + buflen, line, len);
			buflen += len;
		} else {
			memcpy(buffer, line, len);
			buflen = len;
		}
		buffer[buflen] = '\0';
	}
	if (buffer == NULL) {
		return NULL;
	}
	pkinit_debug(mcontext, 3, "Finished reading a \"%s\": \n%.*s\n", type,
		     buflen, buffer);
	/* Check if there are headers -- if we have a blank line to separate
	 * them from the content, we'll treat the first block as headers. */
	headers = (strstr(buffer, "\n\n") != NULL);
	disp = NULL;
	alg = NULL;
	params = NULL;
	if (headers) {
		p = buffer;
		while (p != NULL) {
			/* Find the end of the line. */
			q = p + strcspn(p, "\n");
			if (p == q) {
				break;
			}
			if (strncmp(p, "Proc-Type:", 10) == 0) {
				p += 10;
				p += strspn(p, WHITESPACE);
				tmp = NULL;
				version = strtol(p, &tmp, 10);
				if ((tmp != NULL) && (*tmp == ',')) {
					if (version == 4) {
						disp = PORT_ArenaZAlloc(pool,
									q - tmp);
						memcpy(disp, tmp + 1,
						       q - tmp - 1);
						disp[q - tmp - 1] = '\0';
						pkinit_debug(mcontext, 3,
							     "Data is %s.\n",
							     disp);
					}
				}
			}
			if (strncmp(p, "DEK-Info:", 9) == 0) {
				p += 9;
				p += strspn(p, WHITESPACE);
				tmp = strchr(p, ',');
				if (tmp != NULL) {
					alg = PORT_ArenaZAlloc(pool,
							       tmp - p + 1);
					memcpy(alg, p, tmp - p);
					alg[tmp - p] = '\0';
					params = PORT_ArenaZAlloc(pool,
								  q - tmp);
					memcpy(params, tmp + 1, q - tmp - 1);
					params[q - tmp - 1] = '\0';
					pkinit_debug(mcontext, 3,
						     "Algorithm is %s, "
						     "parameters are \"%s\".\n",
						     alg, params);
				}
			}
			/* Find the next line. */ 
			q = p + strcspn(p, "\n");
			if (*q == '\0') {
				break;
			}
			p = q + 1;
		}
		raw = p + strspn(p, "\n");
	} else {
		raw = buffer + strspn(buffer, "\n");
	}
	pkinit_debug(mcontext, 3, "Raw data is:\n%s\n", raw);
	item = PORT_ArenaZAlloc(pool, sizeof(*item));
	if (item != NULL) {
		item->item_type = type;
		len = strlen(raw);
		item->item = SECITEM_AllocItem(pool, NULL, len);
		if (item->item != NULL) {
			item->item->len = unb64((unsigned char *) raw, len,
						item->item->data);
			if ((disp != NULL) &&
			    (strcmp(disp, "ENCRYPTED") == 0)) {
				if (aa_item_in_ciphertext_list(bag->encrypted_items,
							       item->item,
							       &tmpitem)) {
					pkinit_debug(mcontext, 2,
						     "Not redecrypting a %s, "
						     "using cached copy.\n",
						     tmpitem->item_type);
					PORT_Free(buffer);
					return aa_item_copy(mcontext, pool,
							    tmpitem);
				}
				plaintext = decrypt_item(mcontext, pool,
							 alg, params,
							 item->item,
							 pwcb,
							 pwcb_args);
				if (plaintext == NULL) {
					pkinit_debug(mcontext, 1,
						     "Error decrypting %s.\n",
						     item->item_type);
					PORT_Free(buffer);
					return NULL;
				}
				item->ciphertext = item->item;
				item->item = plaintext;
				tmpitem = aa_item_copy(mcontext, pool, item);
				if (tmpitem != NULL) {
					tmpitem->next = bag->encrypted_items;
					bag->encrypted_items = tmpitem;
				}
			}
		}
	}

	PORT_Free(buffer);
	return item;
}

struct aa_bag *
aa_bag_new(struct module_context *mcontext,
	   CERTCertDBHandle *certdb,
	   PK11PasswordFunc getpw)
{
	struct aa_bag *bag;
	PLArenaPool *pool;

	pool = PORT_NewArena(sizeof(double));
	if (pool == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return NULL;
	}
	bag = PORT_ArenaZAlloc(pool, sizeof(*bag));
	if (bag == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		PORT_FreeArena(pool, PR_TRUE);
		return NULL;
	}
	bag->mcontext = mcontext;
	bag->bag_pool = pool;
	bag->pwcb = getpw;
	bag->certdb = certdb;
	bag->klist = SECKEY_NewPrivateKeyList();
	bag->clist = CERT_NewCertList();
	return bag;
}

void
aa_bag_free(struct aa_bag *bag)
{
	SECKEYPrivateKeyListNode *node;
	PLArenaPool *pool;
	pool = bag->bag_pool;
	for (node = PRIVKEY_LIST_HEAD(bag->klist);
	     !PRIVKEY_LIST_END(node, bag->klist);
	     node = PRIVKEY_LIST_NEXT(node)) {
		SECKEY_DestroyPrivateKey(node->key);
	}
	SECKEY_DestroyPrivateKeyList(bag->klist);
	CERT_DestroyCertList(bag->clist);
	memset(bag, 0, sizeof(*bag));
	PORT_FreeArena(pool, PR_TRUE);
}

static void
decompose_item(struct aa_bag *unused_bag, struct aa_item *unused_item)
{
}

static struct aa_item *
aa_bag_items_from_file(struct aa_bag *bag,
		       const char *path,
		       struct pkinit_pwcb_args *pwcb_args,
		       const char *expected_type, ...)
{
	struct aa_item *item, *results, **next;
	struct pkinit_pwcb_args my_args, *args;
	int i, len;
	FILE *fp;
	char *name;

	fp = fopen(path, "r");
	if (fp == NULL) {
		pkinit_debug(bag->mcontext, 0,
			     "Error reading file \"%s\": %s.\n",
			     path, strerror(errno));
		return NULL;
	}

	results = NULL;
	next = &results;
	i = 0;
	len = strlen(path);
	while (!feof(fp)) {
		if (pwcb_args) {
			my_args = *pwcb_args;
			args = &my_args;
			name = PORT_ArenaZAlloc(bag->bag_pool, len + 32);
			if (name != NULL) {
				snprintf(name, len + 32, "%s:%d", path, i + 1);
				args->name = name;
			}
		} else {
			args = NULL;
		}
		item = aa_item_from_fp(bag->mcontext, bag->bag_pool, bag, fp,
				       bag->pwcb, args);
		if (item != NULL) {
			if ((expected_type != NULL) &&
			    (strcmp(expected_type, item->item_type) != 0)) {
				continue;
			}
			decompose_item(bag, item);
		}
		*next = item;
		while (*next) {
			next = &((*next)->next);
			i++;
		}
	}
	fclose(fp);
	pkinit_debug(bag->mcontext, 3, "Read %d %s items from %s.\n",
		     i, expected_type, path);

	return results;
}

static void
aabag_key_add_to_list(PLArenaPool *pool, struct aa_key_list **list,
		      SECItem *item)
{
	struct aa_key_list *node;
	node = PORT_ArenaZAlloc(pool, sizeof(*node));
	if (node != NULL) {
		node->next = *list;
		node->item = item;
		*list = node;
	}
}

static PRBool
aabag_key_in_list(struct aa_key_list *list, SECItem *item)
{
	while (list != NULL) {
		if (SECITEM_ItemsAreEqual(list->item, item)) {
			return PR_TRUE;
		}
		list = list->next;
	}
	return PR_FALSE;
}

SECStatus
aa_bag_load_key_file(struct aa_bag *bag, const char *path,
		     struct pkinit_pwcb_args *pwcb_args)
{
	struct aa_item *item;
	char nickname_str[PATH_MAX];
	SECItem nickname;
	SECOidData *oid;
	SECKEYPrivateKeyInfo pki;
	SECKEYPrivateKey *key;
	SECStatus result;
	PK11SlotInfo *slot;
	int i;

	result = SECFailure;
	pkinit_debug(bag->mcontext, 3, "Reading private keys from %s.\n", path);
	item = aa_bag_items_from_file(bag, path, pwcb_args,
				      "RSA PRIVATE KEY", NULL);
	if (item != NULL) {
		result = SECSuccess;
	}
	for (i = 0; item != NULL; item = item->next) {
		oid = NULL;
		slot = NULL;
		i++;
		if (strcmp(item->item_type, "RSA PRIVATE KEY") == 0) {
			oid = SECOID_FindOIDByTag(SEC_OID_PKCS1_RSA_ENCRYPTION);
			slot = PK11_GetBestSlot(CKM_RSA_PKCS, pwcb_args);
		}
		if (oid == NULL) {
			pkinit_debug(bag->mcontext, 0,
				     "Error determining OID for key %s:%d.\n",
				     path, i);
			result = SECFailure;
			PK11_FreeSlot(slot);
			continue;
		}
		if (aabag_key_in_list(bag->key_items, item->item)) {
			pkinit_debug(bag->mcontext, 3,
				     "Already imported key %s:%d.\n",
				     path, i);
			continue;
		}
		if (slot == NULL) {
			pkinit_debug(bag->mcontext, 0,
				     "Error locating key storage slot.\n");
		}
		memset(&pki, 0, sizeof(pki));
		pki.arena = bag->bag_pool;
		pki.algorithm.algorithm = oid->oid;
		pki.privateKey = *item->item;
		nickname.len = snprintf(nickname_str, sizeof(nickname_str),
					"%s:%d", path, i);
		nickname.data = (unsigned char *) nickname_str;
		pkinit_debug(bag->mcontext, 3, "Importing key %s ...",
			     nickname_str);
		key = NULL;
		result = PK11_ImportPrivateKeyInfoAndReturnKey(slot,
							       &pki,
							       &nickname,
							       NULL,
							       PR_FALSE,
							       PR_TRUE,
							       PKINIT_KEY_USAGE_FLAGS,
							       &key,
							       pwcb_args);
		if (result != SECSuccess) {
			pkinit_debug(bag->mcontext, 0,
				     "Error importing key %s.\n",
				     nickname_str);
			pkinit_debug(bag->mcontext, 3, " failed.\n");
			result = SECFailure;
			PK11_FreeSlot(slot);
			continue;
		} else {
			pkinit_debug(bag->mcontext, 3, " succeeded.\n");
			aabag_key_add_to_list(bag->bag_pool, &bag->key_items,
					      item->item);
		}
		if (key == NULL) {
			result = SECFailure;
			PK11_FreeSlot(slot);
			continue;
		}
		SECKEY_AddPrivateKeyToListTail(bag->klist, key);
		PK11_FreeSlot(slot);
	}
	return result;
}

static SECStatus
give_cert_trust(struct aa_bag *bag, CERTCertificate *cert,
		const char *nickname, unsigned int more_trust_bits)
{
	CERTCertTrust trust;
	SECStatus result;

	memset(&trust, 0, sizeof(trust));
	if ((result = CERT_GetCertTrust(cert, &trust)) != SECSuccess) {
		pkinit_debug(bag->mcontext, 3,
			     "Warning: error reading current trust level for "
			     "%s, continuing.\n", nickname);
	} else {
		if (((trust.sslFlags & more_trust_bits) == more_trust_bits) &&
		    ((trust.emailFlags & more_trust_bits) == more_trust_bits) &&
		    ((trust.objectSigningFlags &
		      more_trust_bits) == more_trust_bits)) {
			return SECSuccess;
		}
	}
	trust.sslFlags |= more_trust_bits;
	trust.emailFlags |= more_trust_bits;
	trust.objectSigningFlags |= more_trust_bits;
	if ((result = CERT_ChangeCertTrust(bag->certdb,
					   cert, &trust)) != SECSuccess) {
		pkinit_debug(bag->mcontext, 0,
			     "Error setting new trust level for %s.\n",
			     nickname);
	} else {
		pkinit_debug(bag->mcontext, 3,
			     "Marked certificate for \"%s\" with new trust "
			     "levels.\n", cert->subjectName);
	}
	return result;
}

static PRBool
aabag_cert_in_list(CERTCertList *clist, CERTCertificate *cert)
{
	CERTCertListNode *node;
	for (node = CERT_LIST_HEAD(clist);
	     !CERT_LIST_EMPTY(clist) &&
	     !CERT_LIST_END(node, clist);
	     node = CERT_LIST_NEXT(node)) {
		if (SECITEM_ItemsAreEqual(&cert->derCert,
					  &node->cert->derCert)) {
			return PR_TRUE;
		}
	}
	return PR_FALSE;
}

static SECStatus
aa_bag_load_certs_file(struct aa_bag *bag, const char *path, int force_trust,
		       struct pkinit_pwcb_args *pwcb_args)
{
	SECStatus result;
	CERTCertificate **certs, *cert;
	char nickname_str[PATH_MAX];
	SECItem nickname;
	struct aa_item *item;
	int i;

	result = SECFailure;
	pkinit_debug(bag->mcontext, 3, "Reading certificates from %s.\n", path);
	item = aa_bag_items_from_file(bag, path, pwcb_args,
				      "CERTIFICATE", NULL);
	if (item != NULL) {
		result = SECSuccess;
	}

	for (i = 0; item != NULL; item = item->next) {
		i++;
		nickname.len = snprintf(nickname_str, sizeof(nickname_str),
					"%s:%d", path, i);
		nickname.data = (unsigned char *) nickname_str;
		pkinit_debug(bag->mcontext, 3,
			     "Importing certificate %s (%d bytes).\n",
			     nickname_str, item->item->len);
		certs = NULL;
		result = CERT_ImportCerts(bag->certdb, 0, 1,
					  &item->item, &certs,
					  PR_FALSE, PR_FALSE, nickname_str);
		if (certs != NULL) {
			cert = certs[0];
			pkinit_debug(bag->mcontext, 3,
				     "Decoded certificate %s (\"%s\").\n",
				     nickname_str, cert->subjectName);
			result = give_cert_trust(bag, cert, nickname_str, 
						 force_trust);
			if (result == SECSuccess) {
				if (aabag_cert_in_list(bag->clist, cert)) {
					pkinit_debug(bag->mcontext, 3,
						     "Already imported "
						     "%s.\n", nickname_str);
				} else {
					cert = CERT_DupCertificate(cert);
					CERT_AddCertToListTail(bag->clist,
							       cert);
					pkinit_debug(bag->mcontext, 3,
						     "Imported certificate "
						     "%s.\n", nickname_str);
					i++;
				}
			}
			CERT_DestroyCertArray(certs, 1);
		}
		if (result != SECSuccess) {
			pkinit_debug(bag->mcontext, 0,
				     "Failed to import certificate %s: %s.\n",
				     nickname_str,
				     PR_ErrorToString(result,
				    		      PR_LANGUAGE_I_DEFAULT));
		}
	}
	return result;
}

SECStatus
aa_bag_load_cert_file(struct aa_bag *bag, const char *path,
		      struct pkinit_pwcb_args *pwcb_args)
{
	return aa_bag_load_certs_file(bag, path,
				      PKINIT_CA_VALID_FLAGS,
				      pwcb_args);
}

SECStatus
aa_bag_load_cacert_file(struct aa_bag *bag, const char *path,
		        struct pkinit_pwcb_args *pwcb_args)
{
	return aa_bag_load_certs_file(bag, path,
				      PKINIT_CA_TRUST_FLAGS,
				      pwcb_args);
}

static char **
list_dir(struct aa_bag *bag, const char *dir, const char *pattern)
{
	DIR *d;
	struct dirent *ent;
	char **ret, **tmp;
	int i, slash;

	ret = NULL;
	i = 0;
	if (strlen(dir) > 0) {
		slash = (dir[strlen(dir) - 1] == '/');
	} else {
		slash = 0;
	}
	d = opendir(dir);
	if (d != NULL) {
		for (ent = readdir(d); ent != NULL; ent = readdir(d)) {
			if (fnmatch(pattern, ent->d_name, FNM_PATHNAME) == 0) {
				tmp = PORT_Realloc(ret,
						   sizeof(char*) * (i + 2));
				if (tmp != NULL) {
					memcpy(tmp, ret, i * sizeof(char*));
					tmp[i] = PORT_ArenaZAlloc(bag->bag_pool,
								  strlen(dir) +
								  strlen(ent->d_name) +
								  2);
					if (tmp[i] != NULL) {
						sprintf(tmp[i], "%s%s%s",
							dir, slash ? "" : "/",
							ent->d_name);
						tmp[i + 1] = NULL;
						i++;
					}
					ret = tmp;
				} else {
					PORT_Free(ret);
					ret = NULL;
					break;
				}
			}
		}
		closedir(d);
	}
	if (ret != NULL) {
		tmp = PORT_ArenaZAlloc(bag->bag_pool, sizeof(char*) * (i + 1));
		if (tmp != NULL) {
			memcpy(tmp, ret, sizeof(char*) * (i + 1));
			PORT_Free(ret);
			ret = tmp;
		} else {
			PORT_Free(ret);
			ret = NULL;
		}
	}
	return ret;
}

SECStatus
aa_bag_load_key_dir(struct aa_bag *bag, const char *path,
		    const char *pattern, struct pkinit_pwcb_args *pwcb_args)
{
	SECStatus result;
	char **paths;
	unsigned int i;

	paths = list_dir(bag, path, pattern);
	if (paths != NULL) {
		result = SECSuccess;
		for (i = 0; paths[i] != NULL; i++) {
			if (aa_bag_load_key_file(bag, paths[i],
						 pwcb_args) != SECSuccess) {
				result = SECFailure;
			}
		}
		return result;
	}
	return SECFailure;
}

SECStatus
aa_bag_load_cert_dir(struct aa_bag *bag, const char *path,
		     const char *pattern, struct pkinit_pwcb_args *pwcb_args)
{
	SECStatus result;
	char **paths;
	unsigned int i;

	paths = list_dir(bag, path, pattern);
	if (paths != NULL) {
		result = SECSuccess;
		for (i = 0; paths[i] != NULL; i++) {
			if (aa_bag_load_certs_file(bag, paths[i],
						   0,
						   pwcb_args) != SECSuccess) {
				result = SECFailure;
			}
		}
		return result;
	}
	return SECFailure;
}

SECStatus
aa_bag_load_cacert_dir(struct aa_bag *bag, const char *path,
		       const char *pattern, struct pkinit_pwcb_args *pwcb_args)
{
	SECStatus result;
	char **paths;
	unsigned int i;

	paths = list_dir(bag, path, pattern);
	if (paths != NULL) {
		result = SECSuccess;
		for (i = 0; paths[i] != NULL; i++) {
			if (aa_bag_load_certs_file(bag, paths[i],
						   PKINIT_CA_TRUST_FLAGS,
						   pwcb_args) != SECSuccess) {
				result = SECFailure;
			}
		}
		return result;
	}
	return SECFailure;
}

SECKEYPrivateKeyList *
aa_bag_list_loaded_keys(struct aa_bag *bag)
{
	SECKEYPrivateKey *key;
	SECKEYPrivateKeyList *ret;
	SECKEYPrivateKeyListNode *node;
	int i;

	ret = NULL;
	i = 0;
	for (node = PRIVKEY_LIST_HEAD(bag->klist);
	     !PRIVKEY_LIST_END(node, bag->klist);
	     node = PRIVKEY_LIST_NEXT(node)) {
		if (ret == NULL) {
			ret = SECKEY_NewPrivateKeyList();
		}
		key = SECKEY_CopyPrivateKey(node->key);
		SECKEY_AddPrivateKeyToListTail(ret, key);
		i++;
	}
	pkinit_debug(bag->mcontext, 3, "Returning list of %d keys.\n", i);
	return ret;
}

CERTCertList *
aa_bag_list_loaded_certs(struct aa_bag *bag)
{
	CERTCertList *ret;
	CERTCertificate *cert;
	CERTCertListNode *node;
	int i;

	ret = NULL;
	i = 0;
	for (node = CERT_LIST_HEAD(bag->clist);
	     !CERT_LIST_EMPTY(bag->clist) &&
	     !CERT_LIST_END(node, bag->clist);
	     node = CERT_LIST_NEXT(node)) {
		if (ret == NULL) {
			ret = CERT_NewCertList();
		}
		cert = CERT_DupCertificate(node->cert);
		CERT_AddCertToListTail(ret, cert);
		i++;
	}
	pkinit_debug(bag->mcontext, 3, "Returning list of %d certs.\n", i);
	return ret;
}

SECStatus
aa_bag_add_cert_search_dir(struct aa_bag *bag, const char *path)
{
	struct aa_dir *dir, **next;
	next = &bag->cert_dirs;
	while ((*next) != NULL) {
		if (strcmp(path, (*next)->path) == 0) {
			return SECSuccess;
		}
		next = &((*next)->next);
	}
	dir = PORT_ArenaZAlloc(bag->bag_pool, sizeof(*dir));
	if (dir == NULL) {
		pkinit_debug(bag->mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	dir->path = PORT_ArenaStrdup(bag->bag_pool, path);
	if (dir->path == NULL) {
		pkinit_debug(bag->mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	*next = dir;
	return SECSuccess;
}

SECStatus
aa_bag_add_cacert_search_dir(struct aa_bag *bag, const char *path)
{
	struct aa_dir *dir, **next;
	next = &bag->cacert_dirs;
	while ((*next) != NULL) {
		if (strcmp(path, (*next)->path) == 0) {
			return SECSuccess;
		}
		next = &((*next)->next);
	}
	dir = PORT_ArenaZAlloc(bag->bag_pool, sizeof(*dir));
	if (dir == NULL) {
		pkinit_debug(bag->mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	dir->path = PORT_ArenaStrdup(bag->bag_pool, path);
	if (dir->path == NULL) {
		pkinit_debug(bag->mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	*next = dir;
	return SECSuccess;
}

CERTCertList *
aa_bag_find_cert_by_subject(struct aa_bag *bag,
			    SECItem *der_subject,
			    struct pkinit_pwcb_args *pwcb_args)
{
	char pattern[16];
	unsigned char buf[16];
	struct aa_dir *dir;
	char **paths;
	CERTCertificate *cert;
	CERTCertList *clist, *tmplist;
	CERTCertListNode *node;
	unsigned int i;

	if (PK11_HashBuf(SEC_OID_MD5, buf,
			 der_subject->data,
			 der_subject->len) == SECSuccess) {
		sprintf(pattern, "%02x%02x%02x%02x.[0-9]*",
			buf[3], buf[2], buf[1], buf[0]);
		for (dir = bag->cacert_dirs; dir != NULL; dir = dir->next) {
			paths = list_dir(bag, dir->path, pattern);
			if (paths != NULL) {
				for (i = 0; paths[i] != NULL; i++) {
					pkinit_debug(bag->mcontext, 2,
						     "Loading CA certs "
						     "from %s.\n",
						     paths[i]);
					aa_bag_load_cacert_file(bag,
								paths[i],
								pwcb_args);
				}
			}
		}
		for (dir = bag->cert_dirs; dir != NULL; dir = dir->next) {
			paths = list_dir(bag, dir->path, pattern);
			if (paths != NULL) {
				for (i = 0; paths[i] != NULL; i++) {
					pkinit_debug(bag->mcontext, 2,
						     "Loading certs from %s.\n",
						     paths[i]);
					aa_bag_load_cert_file(bag,
							      paths[i],
							      pwcb_args);
				}
			}
		}
	}

	tmplist = aa_bag_list_loaded_certs(bag);
	clist = NULL;
	if (tmplist != NULL) {
		for (node = CERT_LIST_HEAD(tmplist);
		     !CERT_LIST_EMPTY(tmplist) &&
		     !CERT_LIST_END(node, tmplist);
		     node = CERT_LIST_NEXT(node)) {
			pkinit_debug(bag->mcontext, 2,
				     "Comparing \"%s\" to requested subject.\n",
				     node->cert->subjectName);
			if (SECITEM_ItemsAreEqual(&node->cert->derSubject,
						  der_subject)) {
				if (clist == NULL) {
					clist = CERT_NewCertList();
				}
				pkinit_debug(bag->mcontext, 2,
					     "Returning \"%s\".\n",
					     node->cert->subjectName);
				cert = CERT_DupCertificate(node->cert);
				CERT_AddCertToListTail(clist, cert);
			}
		}
		CERT_DestroyCertList(tmplist);
	}
	return clist;
}

CERTCertList *
aa_bag_find_cert_by_subject_key_id(struct aa_bag *bag,
				   SECItem *der_key_id)
{
	CERTCertList *clist, *tmplist;
	CERTCertListNode *node;
	CERTCertificate *cert;
	tmplist = aa_bag_list_loaded_certs(bag);
	clist = NULL;
	if (tmplist != NULL) {
		for (node = CERT_LIST_HEAD(tmplist);
		     !CERT_LIST_EMPTY(tmplist) &&
		     !CERT_LIST_END(node, tmplist);
		     node = CERT_LIST_NEXT(node)) {
			pkinit_debug(bag->mcontext, 2,
				     "Comparing \"%s\" to requested "
				     "subjectKeyId.\n",
				     node->cert->subjectName);
			if (SECITEM_ItemsAreEqual(&node->cert->subjectKeyID,
						  der_key_id)) {
				if (clist == NULL) {
					clist = CERT_NewCertList();
				}
				pkinit_debug(bag->mcontext, 2,
					     "Returning \"%s\".\n",
					     node->cert->subjectName);
				cert = CERT_DupCertificate(node->cert);
				CERT_AddCertToListTail(clist, cert);
			}
		}
		CERT_DestroyCertList(tmplist);
	}
	return clist;
}

CERTCertList *
aa_bag_find_cert_by_issuer_and_serial(struct aa_bag *bag,
				      SECItem *der_issuer,
				      SECItem *der_serial)
{
	CERTCertList *clist, *tmplist;
	CERTCertListNode *node;
	CERTCertificate *cert;
	tmplist = aa_bag_list_loaded_certs(bag);
	clist = NULL;
	if (tmplist != NULL) {
		for (node = CERT_LIST_HEAD(tmplist);
		     !CERT_LIST_EMPTY(tmplist) &&
		     !CERT_LIST_END(node, tmplist);
		     node = CERT_LIST_NEXT(node)) {
			pkinit_debug(bag->mcontext, 2,
				     "Comparing \"%s\" to requested "
				     "issuer and serial.\n",
				     node->cert->subjectName);
			if (SECITEM_ItemsAreEqual(&node->cert->derIssuer,
						  der_issuer) &&
			    SECITEM_ItemsAreEqual(&node->cert->derIssuer,
						  der_serial)) {
				if (clist == NULL) {
					clist = CERT_NewCertList();
				}
				pkinit_debug(bag->mcontext, 2,
					     "Returning \"%s\".\n",
					     node->cert->subjectName);
				cert = CERT_DupCertificate(node->cert);
				CERT_AddCertToListTail(clist, cert);
			}
		}
		CERT_DestroyCertList(tmplist);
	}
	return clist;
}
