/*
 * Copyright (C) 2007 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: bpk5.c,v 1.2 2007/06/20 12:11:19 nalin Exp $"

#include "../config.h"

#include <nspr.h>
#include <nss.h>
#include <certt.h>
#include <pk11pub.h>

#include "bpk5.h"

/* This is PKCS#5 PBKDF1 iff keylen is less that the size of the hash output. */
SECStatus
bpk5_pbkdf1(struct module_context *mcontext, PLArenaPool *pool,
	    const SECItem *password, const SECItem *salt, int iterations,
	    SECOidTag hash, int keylen, SECItem **key)
{
	unsigned int outlen, actual_outlen;
	unsigned char output[64];
	int i, j, parts;
	SECItem *result;
	PK11Context *ctx;

	switch (hash) {
	case SEC_OID_MD2:
	case SEC_OID_MD4:
	case SEC_OID_MD5:
		outlen = 16;
		break;
	case SEC_OID_SHA1:
		outlen = 20;
		break;
	case SEC_OID_SHA256:
		outlen = 32;
		break;
	case SEC_OID_SHA384:
		outlen = 48;
		break;
	case SEC_OID_SHA512:
		outlen = 64;
		break;
	default:
		pkinit_debug(mcontext, 0, "Unsupported hash algorithm.\n");
		return SECFailure;
		break;
	}
	parts = (keylen + outlen - 1) / outlen;
	result = SECITEM_AllocItem(pool, NULL, outlen * parts);
	if (result == NULL) {
		pkinit_debug(mcontext, 0, "Out of memory.\n");
		return SECFailure;
	}
	for (i = 0; i < parts; i++) {
		for (j = 0; j < iterations; j++) {
			ctx = PK11_CreateDigestContext(hash);
			if (ctx == NULL) {
				pkinit_debug(mcontext, 0,
					     "Error beginning digest.\n");
				return SECFailure;
			}
			if ((i > 0) || (j > 0)) {
				if (PK11_DigestOp(ctx, output,
						  outlen) != SECSuccess) {
					pkinit_debug(mcontext, 0,
						     "Error digesting previous "
						     "segment.\n");
					PK11_DestroyContext(ctx, PR_TRUE);
					return SECFailure;
				}
			}
			if ((j == 0) && (password != NULL)) {
				if (PK11_DigestOp(ctx, password->data,
						  password->len) != SECSuccess) {
					pkinit_debug(mcontext, 0,
						     "Error digesting "
						     "password.\n");
					PK11_DestroyContext(ctx, PR_TRUE);
					return SECFailure;
				}
			}
			if ((j == 0) && (salt != NULL)) {
				if (PK11_DigestOp(ctx, salt->data,
						  salt->len) != SECSuccess) {
					pkinit_debug(mcontext, 0,
						     "Error digesting salt.\n");
					PK11_DestroyContext(ctx, PR_TRUE);
					return SECFailure;
				}
			}
			actual_outlen = 0;
			PK11_DigestFinal(ctx, output, &actual_outlen, outlen);
			if (actual_outlen != outlen) {
				pkinit_debug(mcontext, 0, "Error digesting.\n");
				PK11_DestroyContext(ctx, PR_TRUE);
				return SECFailure;
			}
			PK11_DestroyContext(ctx, PR_TRUE);
		}
		memcpy(result->data + (i * outlen), output, outlen);
	}
	memset(output, '\0', sizeof(output));
	result->len = keylen;
	*key = result;
	return SECSuccess;
}
