/*
 * Copyright (C) 2007 Red Hat, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 *
 *   Free Software Foundation
 *   51 Franklin Street, Fifth Floor
 *   Boston, MA  02111-1301
 *   USA.
 */

#ident "$Id: get-pkinit-san.c,v 1.2 2007/05/31 19:08:20 nalin Exp $"

#include "../config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <getopt.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <nss.h>
#include <cert.h>
#include <certt.h>
#include <pk11pub.h>

static void
derdump(SECItem *item)
{
	FILE *pp;
	unsigned int i;
	pp = popen("derdump", "w");
	if (pp) {
		fprintf(stderr, "Dumping %d bytes.\n", item->len);
		for (i = 0; i < item->len; i++) {
			fputc(item->data[i] & 0xff, pp);
		}
		pclose(pp);
	}
}

struct other_name {
	SECItem other_type, other_value;
};

static SEC_ASN1Template
other_name_real_template[] = {
	{
		.kind = SEC_ASN1_SEQUENCE,
		.offset = 0,
		.sub = NULL,
		.size = sizeof(struct other_name),
	},
	{
		.kind = SEC_ASN1_OBJECT_ID,
		.offset = offsetof(struct other_name, other_type),
		.sub = SEC_ObjectIDTemplate,
		.size = sizeof(SECItem),
	},
	{
		.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
			SEC_ASN1_CONSTRUCTED |
			SEC_ASN1_EXPLICIT,
		.offset = offsetof(struct other_name, other_value),
		.sub = SEC_AnyTemplate,
		.size = sizeof(SECItem),
	},
	{
		.kind = 0,
		.offset = 0,
		.sub = NULL,
		.size = 0,
	},
};
static SEC_ASN1Template
other_name_template[] = {
	{
		.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
			SEC_ASN1_CONSTRUCTED,
		.sub = other_name_real_template,
		.size = 0,
	},
};

static SEC_ASN1Template
sequence_of_any_template[] = {
	{
		.kind = SEC_ASN1_SEQUENCE_OF,
		.offset = 0,
		.sub = SEC_AnyTemplate,
		.size = 0,
	},
};

static const SEC_ASN1Template
kerberos_string_template[] = {
	{
	.kind = SEC_ASN1_GENERAL_STRING,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(SECItem),
	},
};

struct realm {
	SECItem name;
};

static const SEC_ASN1Template
realm_template[] = {
	{
	.kind = SEC_ASN1_GENERAL_STRING,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(SECItem),
	},
};

static const SEC_ASN1Template
sequence_of_kerberos_string_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE_OF,
	.offset = 0,
	.sub = &kerberos_string_template,
	.size = 0,
	},
};

struct principal_name {
	SECItem name_type;
	SECItem **name_string;
};

static const SEC_ASN1Template
principal_name_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct principal_name),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct principal_name, name_type),
	.sub = &SEC_IntegerTemplate,
	.size = sizeof(SECItem),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct principal_name, name_string),
	.sub = sequence_of_kerberos_string_template,
	.size = sizeof(struct SECItem**),
	},
	{0, 0, NULL, 0},
};

struct kerberos_principal_name {
	struct realm realm;
	struct principal_name principal_name;
};

const SEC_ASN1Template
kerberos_principal_name_template[] = {
	{
	.kind = SEC_ASN1_SEQUENCE,
	.offset = 0,
	.sub = NULL,
	.size = sizeof(struct kerberos_principal_name),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 0 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct kerberos_principal_name, realm),
	.sub = &realm_template,
	.size = sizeof(struct realm),
	},
	{
	.kind = SEC_ASN1_CONTEXT_SPECIFIC | 1 |
		SEC_ASN1_CONSTRUCTED |
		SEC_ASN1_EXPLICIT,
	.offset = offsetof(struct kerberos_principal_name, principal_name),
	.sub = &principal_name_template,
	.size = sizeof(struct principal_name),
	},
	{0, 0, NULL, 0},
};

static char **
upns_from_other_names(struct other_name **other_names)
{
	unsigned char upn_oid_bytes[] = {0x2b, 0x06, 0x01, 0x04, 0x01,
					 0x82, 0x37, 0x14, 0x02, 0x03};
	SECItem upn_oid = {
		.data = upn_oid_bytes,
		.len = 10,
	};
	int i, n_ret;
	char **ret;
	SECItem name;

	ret = NULL;
	n_ret = 0;
	for (i = 0; (other_names != NULL) && (other_names[i] != NULL); i++) {
		if (SECITEM_ItemsAreEqual(&other_names[i]->other_type,
					  &upn_oid)) {
			if (SEC_ASN1DecodeItem(NULL, &name,
					       SEC_UTF8StringTemplate,
					       &other_names[i]->other_value) ==
			    SECSuccess) {
				ret = PORT_Realloc(ret, sizeof(char *) * (n_ret + 2));
				if (ret) {
					ret[n_ret] = PORT_Alloc(name.len + 1);
					if (ret[n_ret]) {
						memcpy(ret[n_ret], name.data,
						       name.len);
						ret[n_ret][name.len] = '\0';
						n_ret++;
						ret[n_ret] = NULL;
					} else {
						ret = NULL;
						n_ret = 0;
					}
				} else {
					n_ret = 0;
				}
			}
		}
	}

	return ret;
}

static char **
pkinit_from_other_names(struct other_name **other_names)
{
	unsigned char pkinit_oid_bytes[] = {0x2b, 0x06, 0x01, 0x05, 0x02, 0x02};
	SECItem pkinit_oid = {
		.data = pkinit_oid_bytes,
		.len = 6,
	};
	struct kerberos_principal_name pkinit;
	int i, j, k, len, n_ret;
	char **ret;

	ret = NULL;
	n_ret = 0;
	for (i = 0; (other_names != NULL) && (other_names[i] != NULL); i++) {
		if (SECITEM_ItemsAreEqual(&other_names[i]->other_type,
					  &pkinit_oid)) {
			memset(&pkinit, 0, sizeof(pkinit));
			if (SEC_ASN1DecodeItem(NULL, &pkinit,
					       kerberos_principal_name_template,
					       &other_names[i]->other_value) ==
			    SECSuccess) {
				len = pkinit.realm.name.len + 2;
				for (j = 0;
				     (pkinit.principal_name.name_string != NULL) &&
				     (pkinit.principal_name.name_string[j] != NULL);
				     j++) {
					len += (pkinit.principal_name.name_string[j]->len + 1);
				}
				ret = PORT_Realloc(ret, sizeof(char *) * (n_ret + 2));
				if (ret == NULL) {
					n_ret = 0;
				} else {
					ret[n_ret] = PORT_Alloc(len);
					if (ret[n_ret] == NULL) {
						ret = NULL;
						n_ret = 0;
					} else {
						k = 0;
						for (j = 0;
						     (pkinit.principal_name.name_string != NULL) &&
						     (pkinit.principal_name.name_string[j] != NULL);
						     j++) {
							if (j > 0) {
								ret[n_ret][k++] = '/';
							}
							memcpy(ret[n_ret] + k,
							       pkinit.principal_name.name_string[j]->data,
							       pkinit.principal_name.name_string[j]->len);
							k += pkinit.principal_name.name_string[j]->len;
						}
						ret[n_ret][k++] = '@';
						memcpy(ret[n_ret] + k,
						       pkinit.realm.name.data,
						       pkinit.realm.name.len);
						k += pkinit.realm.name.len;
						ret[n_ret][k++] = '\0';
						n_ret++;
						ret[n_ret] = NULL;
					}
				}
			}
		}
	}

	return ret;
}

static struct other_name **
other_names_from_certs(CERTCertificate **cert, int n_certs)
{
	int i, j, n_ret;
	SECItem item, **names;
	struct other_name other, **ret;
	PLArenaPool *pool;

	ret = NULL;
	n_ret = 0;

	pool = PORT_NewArena(sizeof(double));
	for (i = 0; i < n_certs; i++) {
		memset(&item, 0, sizeof(item));
		if (CERT_FindCertExtension(cert[i],
					   SEC_OID_X509_SUBJECT_ALT_NAME,
					   &item) == SECSuccess) {
			names = NULL;
			if (SEC_ASN1DecodeItem(pool, &names,
					       sequence_of_any_template,
					       &item) == SECSuccess) {
				for (j = 0; names[j] != NULL; j++) {
					memset(&other, 0, sizeof(other));
					if (SEC_ASN1DecodeItem(pool, &other,
							       other_name_template,
							       names[j]) == SECSuccess) {
						ret = PORT_Realloc(ret, sizeof(struct other_name *) * (n_ret + 2));
						if (ret != NULL) {
							ret[n_ret] = PORT_Alloc(sizeof(struct other_name));
							if (ret[n_ret] == NULL) {
								PORT_Free(ret);
								ret = NULL;
								n_ret = 0;
							} else {
								ret[n_ret]->other_type = other.other_type;
								ret[n_ret]->other_value = other.other_value;
								n_ret++;
								ret[n_ret] = NULL;
							}
						} else {
							n_ret = 0;
						}
					}
				}
			}
		}
	}
	return ret;
}

static void
cert_der_from_packages(SECItem **packages, int n_packages,
		       SECItem ***der_certs, int *n_der_certs)
{
	SECItem **ret;
	CERTCertificate *cert;
	int i, j;

	ret = NULL;
	j = 0;
	for (i = 0; i < n_packages; i++) {
		cert = CERT_DecodeCertFromPackage((char*) packages[i]->data,
						  packages[i]->len);
		if (cert != NULL) {
			ret = PORT_Realloc(ret, (j + 1) * sizeof(SECItem*));
			if (ret != NULL) {
				ret[j] = SECITEM_AllocItem(NULL, NULL,
							   cert->derCert.len);
				if (ret[j] != NULL) {
					memcpy(ret[j]->data,
					       cert->derCert.data,
					       cert->derCert.len);
				}
				j++;
			} else {
				j = 0;
			}
			CERT_DestroyCertificate(cert);
		}
	}
	*der_certs = ret;
	*n_der_certs = j;
}

static void
cert_packages_from_files(char **file, int count, SECItem ***ret, int *ret_count)
{
	SECItem *item, **items, **tmp;
	struct stat st;
	ssize_t l;
	int c, i, item_count;
	FILE *fp;
	char *p, *q;

	item = NULL;
	items = NULL;
	item_count = 0;
	for (i = 0; i < count; i++) {
		/* Read the file into an item. */
		item = NULL;
		fp = fopen(file[i], "r");
		if (fp != NULL) {
			if (fstat(fileno(fp), &st) != -1) {
				item = SECITEM_AllocItem(NULL, NULL,
							 st.st_size + 1);
				if (item != NULL) {
					for (l = 0; l < st.st_size; l++) {
						c = getc(fp);
						if (c == -1) {
							break;
						}
						item->data[l] = c;
					}
					item->data[l] = '\0';
					if (l != st.st_size) {
						SECITEM_FreeItem(item, PR_TRUE);
						item = NULL;
					}
				}
			}
			fclose(fp);
		}
		if (item == NULL) {
			continue;
		}
		/* Extract each certificate from the file. */
		p = (char *)item->data;
		while ((p != NULL) && (*p != '\0')) {
			p = strstr((char *)p,
				   "-----BEGIN CERTIFICATE-----");
			q = NULL;
			if (p != NULL) {
				q = strstr((char *)p,
					   "-----END CERTIFICATE-----");
			}
			if ((p != NULL) && (q != NULL)) {
				tmp = PORT_Realloc(items,
						   (item_count + 1) *
						   sizeof(SECItem*));
				if (tmp != NULL) {
					items = tmp;
					l = (q - p) +
					    strlen("-----END CERTIFICATE-----");
					items[item_count] = SECITEM_AllocItem(NULL,
									      NULL,
									      l + 1);
					if (items[item_count] != NULL) {
						memmove(items[item_count]->data,
							p, l);
					}
					item_count++;
				} else {
					PORT_Free(items);
					items = NULL;
					item_count = 0;
				}
			}
			p = q;
		}
	}
	*ret = items;
	*ret_count = item_count;
}

int
main(int argc, char **argv)
{
	int i, upn, n_packages, n_der_cert;
	SECItem **packages, **der_cert;
	struct other_name **other_names;
	CERTCertificate **certs;
	char **names;

	upn = 0;
	while ((i = getopt(argc, argv, "uh")) != -1) {
		switch (i) {
		case 'u':
			upn++;
			upn %= 2;
			break;
		case 'h':
		case '-':
			printf("Usage: pkinit-get-san [options] "
			       "[file [...]]\n");
			printf("Options:\n   -u\t"
			       "Show userPrincipalName values instead of "
			       "KerberosPrincipalName values\n");
			return 1;
			break;
		}
	}
	if (optind == argc) {
		fprintf(stderr,
			"Usage: pkinit-get-san [options] "
			"[file [...]]\n");
		fprintf(stderr,
			"Options:\n   -u\t"
			"Show userPrincipalName values instead of "
			"KerberosPrincipalName values\n");
		return 1;
	}

	if (NSS_NoDB_Init(NULL) != SECSuccess) {
		fprintf(stderr, "Error initializing NSS.\n");
		return 1;
	}

	packages = NULL;
	n_packages = 0;
	if (argc - optind > 0) {
		cert_packages_from_files(argv + optind, argc - optind,
					 &packages, &n_packages);
	}

	der_cert = NULL;
	n_der_cert = 0;
	if (n_packages > 0) {
		cert_der_from_packages(packages, n_packages,
				       &der_cert, &n_der_cert);
	}

	certs = NULL;
	if (n_der_cert > 0) {
		if (CERT_ImportCerts(CERT_GetDefaultCertDB(),
				     0,
				     n_der_cert, der_cert,
				     &certs, PR_FALSE, PR_FALSE,
				     "tmp") == SECSuccess) {
		}
	}

	other_names = NULL;
	if (certs != NULL) {
		other_names = other_names_from_certs(certs, n_der_cert);
		CERT_DestroyCertArray(certs, n_der_cert);
	}

	if (upn) {
		names = upns_from_other_names(other_names);
	} else {
		names = pkinit_from_other_names(other_names);
	}
	for (i = 0; (names != NULL) && (names[i] != NULL); i++) {
		printf("%s\n", names[i]);
	}

	if (NSS_Shutdown() != SECSuccess) {
		fprintf(stderr, "Error shutting down NSS.\n");
	}
	return 0;
}
